/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file handextend.c
 */

#include <stdio.h>
#include <string.h>

#include "main.h"

#define HANDEXTEND_MESSAGE	(1<<0)
#define HANDEXTEND_MESSAGEACK	(1<<1)
#define HANDEXTEND_ACKNOWLEDGE	(1<<2)

typedef int (* handproc)(int, message_block *);

typedef struct handlist {
  struct handlist	 * n;
  int			a;
  int			r;
  handproc		p;
} handlist;

static handlist * h = NULL;

static void handextend_scan(int r, handlist * i)
{
  while(i) {
    if((i -> r & r) && (poll_block.action == i -> a)) {
      message_block b;
      memcpy(& b, & poll_block, sizeof(b));
      if(i -> p(i -> a, & b)) return;
    }
    i = i -> n;
  }
}

static void handextend_acknowledge(void)
{
  handextend_scan(HANDEXTEND_ACKNOWLEDGE, h);
}

static void handextend_messageack(void)
{
  handextend_scan(HANDEXTEND_MESSAGEACK, h);
}

static void handextend_message(void)
{
  handextend_scan(HANDEXTEND_MESSAGE, h);
}

void handextend_init(void)
{
#ifdef ERASEHAND
  while(h) {
    handlist * j = h -> n;
    rfree(h);
    h = j;
  }
#endif

  main_addhandler(HANDLER_MESSAGE, handextend_message);
  main_addhandler(HANDLER_MESSAGEACK, handextend_messageack);
  main_addhandler(HANDLER_ACKNOWLEDGE, handextend_acknowledge);
}

void handextend_claim(int a, int r, handproc p)
{
  handlist * j;

  j = rmalloc(sizeof(handlist));
  j -> a = a;
  j -> p = p;
  j -> r = r;
  j -> n = h;
  h = j;
}

void handextend_release(int a, int r, handproc p)
{
  handlist * j = NULL;
  handlist * i;

  i = h;
  while((i) && (i -> a != a) && (i -> p != p) && (i -> r != r)) {
    j = i;
    i = i -> n;
  }
  if(i) {
    if(j) {
      j -> n = i -> n;
    } else {
      h = i -> n;
    }
    rfree(i);
  }
}
